/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Andreas Krennmair <ak@synflood.at> wrote this file.  As long as you retain
 * this notice you can do whatever you want with this stuff. If we meet some day, 
 * and you think this stuff is worth it, you can buy me a beer in return.
 *   -- Andreas Krennmair <ak@synflood.at>
 * ----------------------------------------------------------------------------
 */
#include "poperattess.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <stdint.h>
#include <socket.h>
#include <sys/socket.h>

#include <ip4.h>
#include <io.h>
#include <errmsg.h>
#include <scan.h>


static void sig_chld(int signo)
{
	if (signo == SIGCHLD) {
		pid_t pid;
		int stat;
		while ((pid = waitpid(-1, &stat, WNOHANG)) > 0);
	}
}

static void sig_term(int signo)
{
	signo = signo;
	exit(0);
}

static void usage(const char * argv0) {
	errmsg_warn(argv0, ": usage: ", argv0, " [-p <port>] [-b <bindaddr>]\n",
				"\t-p <port>\t\tbind to port <port> (default: 110)\n",
				"\t-b <bindaddr>\t\tbind to address <bindaddr> (default: 0.0.0.0)", NULL);
	exit(EXIT_FAILURE);
}

enum { TAG_SOCKET4 = 1 };

static void handle_accept4(int s)
{
	char client_ip[4];
	uint16_t client_p;
	int fd;

	while ((fd = socket_accept4(s, client_ip, &client_p)) != -1) {
		int pid = fork();
		if (pid == 0) {
			close(s);
			handle_session(fd, client_ip, client_p);
			close(fd);
			exit(0);
		} else if (pid == -1) {
			errmsg_warn("fork failed: ", strerror(errno), NULL);
		}
		close(fd);
	}

	if (errno == EAGAIN)
		io_eagain(s);
}

int main(int argc, char * argv[])
{
	const char *bind_addr = "0.0.0.0";
	int s;
	char bind_ip[4];
	uint16_t port = 110;
	int rc;
	int c;

	do {
		if ((c = getopt(argc, argv, ":p:b:h")) < 0)
			continue;

		switch (c) {
			case ':':
			case '?':
			case 'h':
				usage(argv[0]);
				break;
			case 'p': {
					unsigned int tmpport;
					if (scan_uint(optarg, &tmpport) <= 0 || tmpport == 0 || tmpport >= (1<<16)) {
						errmsg_warn("Error: invalid port `", optarg, "'\n", NULL);
						usage(argv[0]);
					}
					port = tmpport;
				}
				break;
			case 'b':
				bind_addr = optarg;
				break;
			default: {
					char opt[2] = { c, 0 };
					errmsg_warn("Error: unknown option - ", opt, "\n", NULL);
					usage(argv[0]);
					break;
				}
		}
	} while (c != -1);

	signal(SIGCHLD, sig_chld);

	signal(SIGTERM, sig_term);
	signal(SIGINT, sig_term);
	signal(SIGHUP, sig_term);
	signal(SIGQUIT, sig_term);

	scan_ip4(bind_addr, bind_ip);

	s = socket_tcp4();
	if (s < 0) {
		errmsg_warn("Creating socket failed: ", strerror(errno), NULL);
		exit(1);
	}

	rc = socket_bind4_reuse(s, bind_ip, port);
	if (rc != 0) {
		errmsg_warn("Binding to IP ", bind_addr, " failed: ", strerror(errno), NULL);
		exit(1);
	}

	rc = socket_listen(s, SOMAXCONN);
	if (rc != 0) {
		errmsg_warn("listen failed: ", strerror(errno), NULL);
		exit(1);
	}

	io_fd(s);
	io_setcookie(s, (void *) TAG_SOCKET4);
	io_wantread(s);

	for (;;) {
		int tmpfd;

		io_wait();

		while ((tmpfd = io_canread()) != -1) {
			int cookie = (int) io_getcookie(tmpfd);
			if (cookie == TAG_SOCKET4) {
				handle_accept4(s);
			} else {
				errmsg_warn("unknown file descriptor from io_canread()", NULL);
			}
		}
	}

	return 0;
}
